' PBcards.inc, custom control card game control for PB/DLL
' Public Domain by Borje Hagsten, September 2001
' Updated for PB/Win 7.0, November 2002
' 
' Good place for this file is in PB\WINAPI folder, then simply write
' #INCLUDE "PBCARDS.INC" to enable use of the control. See provided
' sample program for "How to use" information. Look through this
' include file to learn more. Add/change whatever you like in code.
'
' Tried several free DLL's, like CARDS32.DLL and QCARD32.DLL, and while
' they are good enough, they're also a bit bloated and means using a
' third-party tool for something that can be built-in in PB. So, I
' decided to write my own include file instead - this is the result.
'
' PBcards.inc provides almost all necessary things one needs for a card
' game, like bitmaps of card faces and card backs. It can also draw
' empty card holders and even has a simple bonus screen procedure. If
' you don't like that one, it can serve as base for your own.. :)
'
' Provided sample program shows basic use of control. It even has a
' complete patience card game, showing one way of writing such a beast.
' Beast? How hard can it be, you ask yourself? I tell you, it's not easy!
' Sample game is among the easiest - click on cards to select pairs,
' click on pile to deal cards. "All" one has to do is to keep track of deck,
' what cards are on top in piles, what pile have become empty, how many
' cards are there left in deck, does clicked card match previously clicked
' card, has card moved to discard pile, did I eat any dinner today, .. etc.
'
' It usually is a real challenge, but also a fun one. One thing is for sure,
' after a while, you get so used to counting cards, you begin to understand
' the stories told about programmers being trown out of casinos here and there,
' for improving their odds by counting cards at the table.. :)

'********************************************************************
' Control specific messages
'********************************************************************
%CRD_REFRESH        = %WM_USER + 10   'redraws control
%CRD_CLEAR          = %WM_USER + 11   'clears control, lParam is redraw flag
%CRD_SETBKCOLOR     = %WM_USER + 100  'sets background color, lParam is redraw flag
%CRD_SETDECKWIPEX   = %WM_USER + 110  'wParam sets area left of card to be wiped when card is drawn, lParam right..
%CRD_SETDECKWIPEY   = %WM_USER + 111  'wParam sets area above card to be wiped when card is drawn, lParam beneath..
%CRD_GETBRUSH       = %WM_USER + 120  'returns control's main brush (default, green)

%CRD_KEYDOWN        = %WM_USER + 201  'some WM_ .. messages, sent separately to parent
%CRD_KEYUP          = %WM_USER + 202  '(reason for this is, they make it easier to use
%CRD_SYSKEYDOWN     = %WM_USER + 203  ' the control in DDT dialogs..)
%CRD_SYSKEYUP       = %WM_USER + 204
%CRD_LBUTTONDBLCLK  = %WM_USER + 210
%CRD_LBUTTONDOWN    = %WM_USER + 211
%CRD_LBUTTONUP      = %WM_USER + 212
%CRD_MOUSEMOVE      = %WM_USER + 213
%CRD_RBUTTONDBLCLK  = %WM_USER + 217
%CRD_RBUTTONDOWN    = %WM_USER + 218
%CRD_RBUTTONUP      = %WM_USER + 219
%CRD_KILLFOCUS      = %WM_USER + 230
%CRD_SETFOCUS       = %WM_USER + 231

'------------------------------------------------------------------------------
' UDT's, suitable for use as arrays, to build a card game
'------------------------------------------------------------------------------
TYPE CardData
  Number    AS LONG
  Face      AS LONG
  Value     AS LONG
  BackNum   AS LONG
  Drag      AS LONG
  FaceDown  AS LONG
  Selected  AS LONG
  Pile      AS LONG
  PosInPile AS LONG
  r         AS RECT '<- can be used with PtInRect API, good for drag&drop checks..
END TYPE

TYPE PileData
  CardCount AS LONG
  TopCard   AS LONG
  Card2(51) AS LONG
  r         AS RECT
END TYPE

'------------------------------------------------------------------------------
' control specific UDT, to hold some important data
'------------------------------------------------------------------------------
TYPE CRDDATA            'Note, all these UDT's can be freely customized to meet own needs
  hWnd         AS LONG  'control's handle
  id           AS LONG  'control's id
  hInst        AS LONG  'control's instance handle
  hParent      AS LONG  'control's parent's handle
  memDC        AS LONG  'internal, memory DC's handle
  hBit         AS LONG  'internal, memory DC's bitmap handle
  hBitFaces    AS LONG  'internal, card face bitmap's handle 
  hBitBacks    AS LONG  'internal, card back bitmap's handle 
  bkColor      AS LONG  'back ground color
  hBrush       AS LONG  'back ground color brush
  hBrushPile   AS LONG  'patterned brush for piles
  DragDC       AS LONG  'internal, separate DC for drawing dragged card
  hDragBit     AS LONG  'internal, bitmap for above drag DC
  DeckWipeL    AS LONG  'area left of card to be wiped at redraw
  DeckWipeR    AS LONG  'area right of card..
  DeckWipeT    AS LONG  'area above card..
  DeckWipeB    AS LONG  'area beneath card..
END TYPE
'------------------------------------------------------------------------------
DECLARE SUB DrawCardInc(BYVAL hWnd AS LONG, tCard AS CardData, BYVAL cRedraw AS LONG)
DECLARE SUB DrawCardHolderInc(BYVAL hWnd AS LONG, r AS RECT, BYVAL style AS LONG)
DECLARE SUB DrawFrame(BYVAL hWnd AS LONG, tCard AS CardData, BYVAL isSel AS LONG, BYVAL fCol AS LONG)
DECLARE SUB DrawBonusScreen(BYVAL hWnd AS LONG)
DECLARE SUB DoDrag(BYVAL hWnd AS LONG, BYVAL x AS LONG, BYVAL y AS LONG)
DECLARE SUB EndDrag(BYVAL hWnd AS LONG)
DECLARE SUB StartDrag(BYVAL hWnd AS LONG, tCard AS CardData)
DECLARE SUB WipeCard(BYVAL hWnd AS LONG, tCard AS CardData, BYVAL cRedraw AS LONG)

DECLARE FUNCTION CreateCardGame(BYVAL hWnd AS LONG, BYVAL CtrlId AS LONG, _
                                BYVAL vLeft AS LONG, BYVAL vTop AS LONG, _
                                BYVAL vWidth AS LONG, BYVAL vHeight AS LONG, _
                                BYVAL wStyle AS LONG, BYVAL wStyleEx AS LONG, _
                                BYVAL DDTstyle AS LONG) AS LONG
DECLARE FUNCTION MakePatternBrush(BYVAL hWnd AS LONG) AS LONG
DECLARE FUNCTION PBCardProc (BYVAL hWnd AS LONG, BYVAL wMsg AS LONG, _
                             BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG

'********************************************************************
' Initialize and create the Game control
'--------------------------------------------------
' hWnd             = Parent's handle
' CtrlId           = Control id
' vLeft, vTop      = position of upper left corner
' vWidth, vHeight  = control's width and height
' wStyle           = Window style, like %WS_CHILD OR %WS_VISIBLE, etc..
' wStyleEX         = Window extra style, like %WS_EX_CLIENTEDGE, etc..
' DDTstyle, %TRUE  = uses DDT to create the control
'           %FALSE = uses CreateWindowEx to create it.
' Reason for this is that they use different coordinate systems
' and some prefer DDT, others SDK.. :-)
'********************************************************************
FUNCTION CreateCardGame(BYVAL hWnd AS LONG, BYVAL CtrlId AS LONG, _
                        BYVAL vLeft AS LONG, BYVAL vTop AS LONG, _
                        BYVAL vWidth AS LONG, BYVAL vHeight AS LONG, _
                        BYVAL wStyle AS LONG, BYVAL wStyleEx AS LONG, _
                        BYVAL DDTstyle AS LONG) AS LONG

  LOCAL hCGame AS LONG, wc AS WNDCLASS, szName AS ASCIIZ * 20

  szName      = "PBCARDGAME"
  IF GetClassInfo(GetModuleHandle(BYVAL %NULL), szName, wc) = 0 THEN 'if not already registered
     wc.style         = %CS_DBLCLKS OR %CS_PARENTDC OR %CS_HREDRAW OR %CS_VREDRAW
     wc.cbWndExtra    = 4                           '4 pre-allocated bytes for pointer to UDT
     wc.lpfnWndProc   = CODEPTR(PBCardProc)
     wc.hInstance     = GetModuleHandle(BYVAL %NULL)
     wc.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
     wc.lpszClassName = VARPTR(szName)

     CALL RegisterClass(wc)
  END IF

  IF DDTstyle THEN 'create control using dialog units
     CONTROL ADD "PBCARDGAME", hWnd, CtrlId, "", vLeft, vTop, vWidth, vHeight, _
                 wStyle, wStyleEx
     CONTROL HANDLE hWnd, CtrlId TO hCGame

  ELSE             'create control using pixels
     hCGame = CreateWindowEx(wStyleEx, "PBCARDGAME", BYVAL 0, wStyle, _
                              vLeft, vTop, vWidth, vHeight, _
                              hwnd, CtrlId, wc.hInstance, BYVAL %NULL)
  END IF

  FUNCTION = hCGame 'return handle
END FUNCTION

'********************************************************************
'  Card Game Procedure
'********************************************************************
FUNCTION PBCardProc (BYVAL hWnd AS LONG, BYVAL wMsg AS LONG, _
                    BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG

  LOCAL crd AS CRDDATA PTR
  IF wMsg <> %WM_CREATE THEN crd = GetWindowLong(hWnd, 0) 'get control-specific UDT data

  SELECT CASE wMsg
     CASE %WM_CREATE
        crd = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, SIZEOF(@crd))
        IF crd THEN
           SetWindowLong hWnd, 0, crd     'Store pointer for later use
        ELSE
           FUNCTION = -1 : EXIT FUNCTION  'Return -1 to break the action on failure
        END IF

        LOCAL ps  AS PAINTSTRUCT
        LOCAL rc  AS RECT
        LOCAL hDC AS LONG, memDC AS LONG, hbit AS LONG

        @crd.hParent = GetWindowLong(hWnd, %GWL_HWNDPARENT)
        @crd.hWnd    = hWnd
        @crd.id      = GetWindowLong(hWnd, %GWL_ID)
        @crd.hInst   = GetWindowLong(hWnd, %GWL_HINSTANCE)

        @crd.bkColor     = RGB(0,128,0)
        @crd.hBrush      = CreateSolidBrush(@crd.bkColor)
        @crd.hBrushPile  = MakePatternBrush(hWnd)

        @crd.hBitFaces   = LoadImage(@crd.hInst, BYVAL 1, %IMAGE_BITMAP, 0, 0, 0) 'card faces
        @crd.hBitBacks   = LoadImage(@crd.hInst, BYVAL 2, %IMAGE_BITMAP, 0, 0, 0) 'card backsides

        SendMessage hWnd, %CRD_CLEAR, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_DESTROY
        DeleteObject @crd.hBitFaces     'clean up memory - card bitmaps
        DeleteObject @crd.hBitBacks                       'card backs
        DeleteObject @crd.hBrush                          'background brush
        DeleteObject @crd.hBrushPile                      'discard pile brush
        DeleteObject SelectObject(@crd.memDC, @crd.hBit)  'memDC's bitmap
        DeleteDC @crd.memDC                               'memDC itself
        IF crd THEN CALL HeapFree(GetProcessHeap(), 0, BYVAL crd)
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_ERASEBKGND
        FUNCTION = 1 : EXIT FUNCTION

     CASE %WM_PAINT                  'This is where the game is drawn.
        hDC = BeginPaint(hWnd, ps)
          GetClientRect hWnd, rc
          BitBlt hDC, 0, 0, rc.nRight, rc.nBottom, @crd.memDC, 0, 0, %SRCCOPY
        EndPaint hWnd, ps
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_GETDLGCODE : FUNCTION = %DLGC_WANTARROWS : EXIT FUNCTION

     '---------------------------------------------------------------
     'send some custom-made messages to parent
     '---------------------------------------------------------------
     CASE %WM_KEYDOWN
        SendMessage GetParent(hWnd), %CRD_KEYDOWN, wParam, lParam

     CASE %WM_KEYUP
        SendMessage GetParent(hWnd), %CRD_KEYUP, wParam, lParam

     CASE %WM_SYSKEYDOWN
        SendMessage GetParent(hWnd), %CRD_SYSKEYDOWN, wParam, lParam

     CASE %WM_SYSKEYUP
        SendMessage GetParent(hWnd), %CRD_SYSKEYUP, wParam, lParam

     CASE %WM_LBUTTONDBLCLK
        IF GetFocus <> hWnd THEN SetFocus hWnd
        SendMessage @crd.hParent, %CRD_LBUTTONDBLCLK, wParam, lParam

     CASE %WM_LBUTTONDOWN
        IF GetFocus <> hWnd THEN SetFocus hWnd
        SendMessage @crd.hParent, %CRD_LBUTTONDOWN, wParam, lParam

     CASE %WM_LBUTTONUP
        SendMessage @crd.hParent, %CRD_LBUTTONUP, wParam, lParam

     CASE %WM_MOUSEMOVE
        SendMessage @crd.hParent, %CRD_MOUSEMOVE, wParam, lParam

     CASE %WM_RBUTTONDBLCLK
        SendMessage @crd.hParent, %CRD_RBUTTONDBLCLK, wParam, lParam

     CASE %WM_RBUTTONDOWN
        SendMessage @crd.hParent, %CRD_RBUTTONDOWN, wParam, lParam

     CASE %WM_RBUTTONUP
        SendMessage @crd.hParent, %CRD_RBUTTONUP, wParam, lParam

     CASE %WM_KILLFOCUS
        SendMessage @crd.hParent, %CRD_KILLFOCUS, wParam, lParam
        
     CASE %WM_SETFOCUS
        SendMessage @crd.hParent, %CRD_SETFOCUS, wParam, lParam

     '---------------------------------------------------------------
     'some own messages, for easy redraw and settings
     '---------------------------------------------------------------
     CASE %CRD_REFRESH
        InvalidateRect hWnd, BYVAL %NULL, 0
        UpdateWindow hWnd
        FUNCTION = 0 : EXIT FUNCTION

     CASE %CRD_CLEAR
        IF @crd.memDC AND @crd.hBit THEN
           DeleteObject SelectObject(@crd.memDC, @crd.hBit)
           DeleteDC @crd.memDC
        END IF

        GetClientRect hWnd, rc
        hDC          = GetDC(hWnd)                                        'get dialog's DC
        @crd.memDC   = CreateCompatibleDC(hDC)                            'create compatible memDC
        @crd.hBit    = CreateCompatibleBitmap(hDC, rc.nRight, rc.nBottom) 'create compatible bitmap
        @crd.hBit    = SelectObject(@crd.memDC, @crd.hBit)                'select bitmap into memDC
        FillRect @crd.memDC, rc, @crd.hBrush                              'paint background color
        ReleaseDC hWnd, hDC
        IF lParam THEN SendMessage hWnd, %CRD_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %CRD_GETBRUSH       : Function = @crd.hBrush : EXIT FUNCTION
     CASE %CRD_SETDECKWIPEX   : @crd.DeckWipeL = wParam : @crd.DeckWipeR = lParam
     CASE %CRD_SETDECKWIPEY   : @crd.DeckWipeT = wParam : @crd.DeckWipeB = lParam

     CASE %CRD_SETBKCOLOR
        @crd.bkColor = wParam                            'wParam holds new color
        DeleteObject @crd.hBrush                         'need to delete old brushes
        DeleteObject @crd.hBrushPile
        @crd.hBrush     = CreateSolidBrush(@crd.bkColor) 'create new brushes
        @crd.hBrushPile = MakePatternBrush(hWnd)
        SendMessage hWnd, %CRD_CLEAR, 0, lParam          're-initiate memDC, lParam decides redraw
        FUNCTION = 0 : EXIT FUNCTION

  END SELECT

  '* Default processing for other messages.
  FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)

END FUNCTION

'
' draw desired card
'
SUB DrawCardInc(BYVAL hWnd AS LONG, tCard AS CardData, BYVAL cRedraw AS LONG)
  LOCAL memDC AS LONG, hBitOld AS LONG, hRgn AS LONG
  LOCAL xv AS LONG, yv AS LONG, crd AS CRDDATA PTR

  crd   = GetWindowLong(hWnd, 0)                'get control-specific UDT data
  memDC = CreateCompatibleDC(@crd.memDC)        'create temporary memDC for bitmap
  hRgn  = CreateRoundRectRgn(tCard.r.nLeft, _   'set region to avoid sharp corners
                             tCard.r.nTop, tCard.r.nRight + 1, tCard.r.nBottom + 1, 6, 6)
  SelectObject @crd.memDC, hRgn

  IF tCard.FaceDown = 0 THEN  'draw card
     hBitOld = SelectObject(memDC, @crd.hBitFaces) 'select card face bitmap into memDC
     xv = (tCard.Value - 1) * 71                   'position of card to draw
     yv = (tCard.Face - 1) * 96
  ELSE                                             'else draw backside
     hBitOld = SelectObject(memDC, @crd.hBitBacks) 'select backside bitmap into memDC
     xv = tCard.BackNum * 71
     yv = 0
  END IF

  IF tCard.Selected = 0 THEN      'normal copy
     BitBlt @crd.memDC, tCard.r.nLeft, tCard.r.nTop, 71, 96, memDC, xv, yv, %SRCCOPY
  ELSE                            'inverted copy, for selected cards
     BitBlt @crd.memDC, tCard.r.nLeft, tCard.r.nTop, 71, 96, memDC, xv, yv, %NOTSRCCOPY
  END IF

  SelectObject memDC, hBitOld     'select proper bitmap back
  DeleteDC memDC                  'delete temporary memDC
  DeleteObject hRgn               'delete region
  SelectClipRgn @crd.memDC, %NULL 'tell control - no clip region

  IF cRedraw = 1 THEN              'if redraw flag - update screen (trigger WM_PAINT)
     RedrawWindow hWnd, ByVal 0, ByVal 0, %RDW_INVALIDATE OR %RDW_UPDATENOW
  END IF

END SUB

'
' start drag procedure - first use WipeCard, the draw card(s) under drag card
'
SUB StartDrag(BYVAL hWnd AS LONG, tCard AS CardData)
  LOCAL hDC AS LONG, memDC AS LONG, hBitOld AS LONG, hRgn AS LONG
  LOCAL xv AS LONG, yv AS LONG, crd AS CRDDATA PTR

  crd   = GetWindowLong(hWnd, 0)                'get control-specific UDT data

  hDC           = GetDC(hWnd)                              'get dialog's DC
  @crd.DragDC   = CreateCompatibleDC(hDC)                  'create compatible DragDC
  @crd.hDragBit = CreateCompatibleBitmap(hDC, 71, 96)      'create compatible bitmap
  @crd.hDragBit = SelectObject(@crd.DragDC, @crd.hDragBit) 'select bitmap into DragDC
  hRgn          = CreateRoundRectRgn(tCard.r.nLeft, _      'set region to avoid sharp corners
                  tCard.r.nTop, tCard.r.nRight + 1, tCard.r.nBottom + 1, 6, 6)
  SelectObject hDC, hRgn

  memDC = CreateCompatibleDC(@crd.DragDC)          'create temporary memDC for bitmap

  IF tCard.FaceDown = 0 THEN  'draw card
     hBitOld = SelectObject(memDC, @crd.hBitFaces) 'select card face bitmap into memDC
     xv = (tCard.Value - 1) * 71                   'position of card to draw
     yv = (tCard.Face - 1) * 96
  ELSE                                             'else draw backside
     hBitOld = SelectObject(memDC, @crd.hBitBacks) 'select backside bitmap into memDC
     xv = tCard.BackNum * 71
     yv = 0
  END IF

  BitBlt @crd.DragDC, 0, 0, 71, 96, memDC, xv, yv, %SRCCOPY
  BitBlt hDC, tCard.r.nLeft, tCard.r.nTop, 71, 96, @crd.DragDC, 0, 0, %SRCCOPY

  SelectObject memDC, hBitOld     'select proper bitmap back
  DeleteDC memDC                  'delete temporary memDC
  DeleteObject hRgn               'delete region
  ReleaseDC hWnd, hDC

END SUB

'
' do drag procedure
'
SUB DoDrag(BYVAL hWnd AS LONG, BYVAL x AS LONG, BYVAL y AS LONG)
  LOCAL I AS LONG, J AS LONG, hDC AS LONG, memDC AS LONG, hBit AS LONG, hRgn AS LONG
  LOCAL rc AS RECT, crd AS CRDDATA PTR

  crd  = GetWindowLong(hWnd, 0)                              'get control-specific UDT data
  GetClientRect hWnd, rc                                     'get control's size
  hDC   = GetDC(hWnd)                                        'get dialog's DC
  memDC = CreateCompatibleDC(hDC)                            'create compatible memDC
  hBit  = CreateCompatibleBitmap(hDC, rc.nRight, rc.nBottom) 'create compatible bitmap
  hBit  = SelectObject(memDC, hBit)                          'select bitmap into memDC
  hRgn  = CreateRoundRectRgn(x, y, x + 72, y + 97, 6, 6)     'use region to avoid sharp corners

  BitBlt memDC, 0, 0, rc.nRight, rc.nBottom, @crd.memDC, 0, 0, %SRCCOPY

  SelectObject memDC, hRgn
  BitBlt memDC, x, y, 71, 96, @crd.DragDC, 0, 0, %SRCCOPY
  DeleteObject hRgn                                           'delete region

  BitBlt hDC, 0, 0, rc.nRight, rc.nBottom, memDC, 0, 0, %SRCCOPY

  DeleteObject SelectObject(memDC, hBit)
  DeleteDC memDC
  ReleaseDC hWnd, hDC

END SUB

'
' end drag procedure
'
SUB EndDrag(BYVAL hWnd AS LONG)
  LOCAL crd AS CRDDATA PTR
  crd    = GetWindowLong(hWnd, 0)                                'get control-specific UDT data

  IF @crd.DragDC AND @crd.hDragBit THEN
     DeleteObject SelectObject(@crd.DragDC, @crd.hDragBit)
     DeleteDC @crd.DragDC
  END IF

END SUB

'
' Wipe out a card - good for use in combination with StartDrag
'
SUB WipeCard(BYVAL hWnd AS LONG, tCard AS CardData, BYVAL cRedraw AS LONG)
  LOCAL hBrush AS LONG, hPen AS LONG, crd AS CRDDATA PTR

  crd    = GetWindowLong(hWnd, 0)                                'get control-specific UDT data
  hBrush = SelectObject(@crd.memDC, @crd.hBrush)                 'select background brush into DC
  hPen   = CreatePen(%PS_SOLID, 1, @crd.bkColor)                 'create neutral pen to wipe frame
  hPen   = SelectObject(@crd.memDC, hPen)                        'select pen into DC

  RoundRect @crd.memDC, tCard.r.nLeft - 1, tCard.r.nTop - 1, _
                 tCard.r.nRight + 2, tCard.r.nBottom + 2, 8, 8   'draw frame same as card shape

  DeleteObject SelectObject(@crd.memDC, hPen)                    'delete pen
  SelectObject @crd.memDC, hBrush                                'delete brush

  IF cRedraw THEN SendMessage hWnd, %CRD_REFRESH, 0, 0
END SUB

'
' Draw a hollow frame around selected card
'
SUB DrawFrame(BYVAL hWnd AS LONG, tCard AS CardData, BYVAL isSel AS LONG, BYVAL fCol AS LONG)
  LOCAL hPen AS LONG, hBrush AS LONG, crd AS CRDDATA PTR

  crd    = GetWindowLong(hWnd, 0)                                'get control-specific UDT data
  hBrush = SelectObject(@crd.memDC, GetStockObject(%NULL_BRUSH)) 'select a hollow brush into DC

  IF isSel THEN
     hPen = CreatePen(%PS_SOLID, 2, fCol)                        'create a yellow pen for frame
  ELSE
     hPen = CreatePen(%PS_SOLID, 2, @crd.bkColor)                'create neutral pen to wipe frame
  END IF
  hPen   = SelectObject(@crd.memDC, hPen)                        'select pen into DC

  RoundRect @crd.memDC, tCard.r.nLeft - 1, tCard.r.nTop - 1, _
                 tCard.r.nRight + 2, tCard.r.nBottom + 2, 8, 8   'draw frame same as card shape

  DeleteObject SelectObject(@crd.memDC, hPen)                    'delete pen
  SelectObject @crd.memDC, hBrush                                'delete brush

  SendMessage hWnd, %CRD_REFRESH, 0, 0
END SUB

'
' Draw card holders
'
SUB DrawCardHolderInc(BYVAL hWnd AS LONG, r AS RECT, BYVAL style AS LONG)
  LOCAL hPen AS LONG, hBrush AS LONG, lb AS LOGBRUSH
  LOCAL crd AS CRDDATA PTR
  crd  = GetWindowLong(hWnd, 0) 'Get stored data
  hPen = CreatePen(%PS_SOLID, 1, RGB(0,0,0))
  hPen = SelectObject(@crd.memDC, hPen)

  IF style THEN 'DRAW PATTERNED HOLDER FOR DECK AND DISCARD PILE
     hBrush  = SelectObject(@crd.memDC, @crd.hBrushPile)
     r.nLeft   = r.nLeft   - @crd.DeckWipeL
     r.nRight  = r.nRight  + @crd.DeckWipeR
     r.nTop    = r.nTop    - @crd.DeckWipeT
     r.nBottom = r.nBottom + @crd.DeckWipeB

     FillRect @crd.memDC, r, @crd.hBrush

     r.nLeft   = r.nLeft   + @crd.DeckWipeL
     r.nRight  = r.nRight  - @crd.DeckWipeR
     r.nTop    = r.nTop    + @crd.DeckWipeT
     r.nBottom = r.nBottom - @crd.DeckWipeB

  ELSE          'DRAW HOLDER FOR ACTIVE PLAYING CARD
     hBrush  = SelectObject(@crd.memDC, @crd.hBrush)
     FillRect @crd.memDC, r, @crd.hBrush
  END IF

  RoundRect @crd.memDC, r.nLeft, r.nTop, r.nRight, r.nBottom, 6, 6  'same as card shape
  IF hPen THEN DeleteObject SelectObject(@crd.memDC, hPen)          'clean up pen

  IF style = 2 THEN        'DRAW GREEN CIRCLE IN EMPTY DECK CARD HOLDER (DEFAULT)
     hPen = CreatePen(%PS_SOLID, 7, RGB(0,255,0))
     hPen = SelectObject(@crd.memDC, hPen)
     Ellipse @crd.memDC, r.nLeft + 10, r.nTop + 22, r.nLeft + 62, r.nTop + 74

  ELSEIF style = 3 THEN    'DRAW RED CROSS IN DECK CARD HOLDER (OPTIONAL)
     lb.lbstyle= %BS_SOLID
     lb.lbcolor= RGB(255,0,0)
     hPen = ExtCreatePen(%PS_GEOMETRIC OR %PS_SOLID OR %PS_ENDCAP_FLAT, 7, lb, 0, BYVAL 0)
     hPen = SelectObject(@crd.memDC, hPen)
     BeginPath @crd.memDC                          'must use path for %PS_ENDCAP_FLAT in win95/98
     MoveTo @crd.memDC, r.nLeft + 9, r.nTop + 21   'move into position
     LineTo @crd.memDC, r.nLeft + 61, r.nTop + 73  'draw lines (build macro for StrokePath)
     MoveTo @crd.memDC, r.nLeft + 9, r.nTop + 73
     LineTo @crd.memDC, r.nLeft + 61, r.nTop + 21
     EndPath @crd.memDC                            'we have a path, so end path
     StrokePath @crd.memDC                         'finally, render path
  END IF

  DeleteObject SelectObject(@crd.memDC, hPen)      'clean up memory
  SelectObject @crd.memDC, hBrush

END SUB

'
' Create patterned brush for card holders - same pattern as MS Solitaire
' (returned brush is deleted in WinMainProc's WM_DESTROY..)
'
FUNCTION MakePatternBrush(BYVAL hWnd AS LONG) AS LONG
  LOCAL hDC AS LONG, hDCmem AS LONG, hBmp AS LONG, hBmpOld AS LONG
  LOCAL crd AS CRDDATA PTR, rc AS RECT

  crd     = GetWindowLong(hWnd, 0)             'get control-specific UDT data
  hDC     = GetDC(hWnd)                        'get dialog's DC
  hDCmem  = CreateCompatibleDC(hDC)            'create compatible memDC
  hBmp    = CreateCompatibleBitmap(hDC, 8, 8)  'create compatible bitmap
  hBmpOld = SelectObject(hDCmem, hBmp)         'select bitmap into memDC
  rc.nright = 8 : rc.nbottom = 8               'bitmap is 8x8 pixels

  FillRect hdcMem, rc, @crd.hBrush             'fill bitmap with dialog's color (green)

  'Bitmap is 8x8 and looks like -->                0 1 2 3 4 5 6 7
                                               ' 0 o x o o o o o o
  'So paint following pixels black               1 o o o o o o o o
  SetPixelV hdcMem, 1, 0, RGB(0,0,0)            ' 2 o o o x o o o x
  SetPixelV hdcMem, 3, 2, RGB(0,0,0)            ' 3 o o o o o o o o
  SetPixelV hdcMem, 3, 6, RGB(0,0,0)            ' 4 o o o o o x o o
  SetPixelV hdcMem, 5, 4, RGB(0,0,0)            ' 5 o o o o o o o o
  SetPixelV hdcMem, 7, 2, RGB(0,0,0)            ' 6 o o o x o o o x
  SetPixelV hdcMem, 7, 6, RGB(0,0,0)            ' 7 o o o o o o o o

  FUNCTION = CreatePatternBrush(hBmp)          'return id for created brush

  SelectObject hDCmem, hBmpOld                 'now we can clean up memory
  DeleteObject hBmp
  DeleteDC hDCmem
  ReleaseDC hWnd, hDC

END FUNCTION

'
' draw bonus screen when/if game is won
'
SUB DrawBonusScreen(BYVAL hWnd AS LONG)
  LOCAL I AS LONG, J AS LONG, xv AS LONG, yv AS LONG
  LOCAL memDC AS LONG, hBitOld AS LONG, hRgn AS LONG
  LOCAL x2 AS SINGLE, y2 AS SINGLE, rc AS RECT, crd AS CRDDATA PTR

  crd = GetWindowLong(hWnd, 0)                  'get control-specific UDT data
  memDC   = CreateCompatibleDC(@crd.memDC)      'need memDC for bitmap
  hBitOld = SelectObject(memDC, @crd.hBitFaces) 'select card face bitmap into memDC
  GetClientRect hWnd, rc                        'get control's size
  rc.nRight  = rc.nRight - 1
  rc.nBottom = rc.nBottom - 1

  x2 = (rc.nRight  - 71) / 12                   'calculate steps
  y2 = (rc.nBottom - 96) / 3

  FOR I = 0 TO 3
     FOR J = 0 TO 12
        xv = J * 71 : yv = I * 96               'position of card to draw
        hRgn = CreateRoundRectRgn(J * x2, I * y2, J * x2 + 72, I * y2 + 97, 6, 6)
        SelectObject @crd.memDC, hRgn

        BitBlt @crd.memDC, J * x2, I * y2, 71, 96, memDC, xv, yv, %SRCCOPY     'normal copy

        DeleteObject hRgn
        RedrawWindow hWnd, ByVal 0, ByVal 0, %RDW_INVALIDATE OR %RDW_UPDATENOW
        SLEEP 15
     NEXT
  NEXT

  SelectClipRgn @crd.memDC, %NULL  'tell control - no clip region
  SelectObject memDC, hBitOld      'select old bitmap back
  DeleteDC memDC                   'delete memDC

  'DIM SCREEN BY PAINTING EVERY OTHER PIXEL
  FOR I = 0 TO rc.nBottom STEP 2
     IF I MOD 2 = 0 THEN
        FOR J = 0 TO rc.nRight STEP 3
           SetPixelV @crd.memDC, J, I, GetSysColor(%COLOR_3DDKSHADOW)
        NEXT
     ELSE
        FOR J = 1 TO rc.nRight STEP 3
           SetPixelV @crd.memDC, J, I, GetSysColor(%COLOR_3DDKSHADOW)
        NEXT
     END IF
  NEXT

  RedrawWindow hWnd, ByVal 0, ByVal 0, %RDW_INVALIDATE OR %RDW_UPDATENOW 'trigger WM_PAINT to update screen

END SUB
